import inspect
import click
import torch
import torch.nn as nn
import dataclasses as dtc

from ..audios.features import MuLawSignal
from ..data import TBPTTSampler, Feature, AsSlice, AsFramedSlice, Input, Target
from .parts import SuperAdam, SequenceModel
from ..networks.sample_rnn import SampleRNNNetwork

from . import model, IData

__all__ = [
    'SampleRNNData',
    'SampleRNN',
    'demo'
]


@dtc.dataclass
class SampleRNNData(IData):
    feature: Feature = None
    batch_size: int = 16
    chunk_len: int = 16000 * 8
    batch_seq_len: int = 512

    @classmethod
    def schema(cls, sr=22050, emphasis=0., q_levels=256):
        return {'qx': MuLawSignal(sr=sr, emphasis=emphasis, q_levels=q_levels)}

    def batch_signature(self, stage='fit'):
        batch_seq_len, frame_sizes = tuple(getattr(self, key) for key in ["batch_seq_len", "frame_sizes"])
        shifts = [frame_sizes[0] - size for size in frame_sizes]
        inputs = []
        for fs, shift in zip(frame_sizes[:-1], shifts[:-1]):
            inputs.append(
                Input('qx', AsFramedSlice(shift, batch_seq_len, frame_size=fs,
                                          as_strided=False)))
        inputs.append(
            Input('qx', AsFramedSlice(shifts[-1], batch_seq_len+frame_sizes[-1]-1, frame_size=frame_sizes[-1],
                                      as_strided=True)))
        targets = Target('qx', AsSlice(shift=frame_sizes[0], length=batch_seq_len))
        if stage in ('fit', 'train', 'val'):
            return inputs, targets
        # test, predict, generate
        return Input('qx', AsSlice(0, batch_seq_len))

    @classmethod
    def dependant_hp(cls, db):
        return dict(feature=db.schema['qx'], q_levels=db.schema['qx'].q_levels)

    def loader_kwargs(self, stage, datamodule):
        ds = datamodule.datasets[stage]
        N = len(ds)
        batch_sampler = TBPTTSampler(N,
                                     self.batch_size,
                                     self.chunk_len,
                                     self.batch_seq_len)
        return dict(batch_sampler=batch_sampler)


@model
class SampleRNN(
    # db schema, batch sig
    SampleRNNData,
    # training step, generate routine & interface
    SequenceModel,
    # optimizer & scheduler
    SuperAdam,
    # nn.Module, generate_ function
    SampleRNNNetwork
):

    def setup(self, stage: str):
        SequenceModel.setup(self, stage)
        SuperAdam.setup(self, stage)

    @staticmethod
    def loss_fn(output, target):
        criterion = nn.CrossEntropyLoss(reduction="mean")
        return {"loss": criterion(output.view(-1, output.size(-1)), target.view(-1))}

    def on_train_batch_start(self, batch, batch_idx, dataloader_idx):
        if (batch_idx * self.batch_seq_len) % self.chunk_len == 0:
            self.reset_h0()

    def encode_inputs(self, inputs: torch.Tensor):
        return self.feature.encode(inputs)

    def decode_outputs(self, outputs: torch.Tensor):
        return self.feature.decode(outputs)


def make_click_command(f):
    aspec = inspect.getfullargspec(f)
    if aspec.defaults is not None:
        for arg, default in zip(reversed(aspec.args), reversed(aspec.defaults)):
            opt = click.option('--' + arg.replace("_", "-"), default=default)
            f = opt(f)
    return click.command()(f)


def demo():
    """### import and arguments"""
    import mimikit as mmk
    import torch

    # DATA

    # list of files or directories to use as data ("./" is the cwd of the notebook)
    sources = ['./data']
    # audio sample rate
    sr = 16000
    # number of quantization levels (256 -> 8-bit)
    q_levels = 256

    # NETWORK

    # how many samples each tier receives as input
    # stick to decreasing sequences, size_i must be divisible by size_i+1 and
    # last 2 numbers must be equal. You can have as many tiers as you want.
    frame_sizes = (16, 4, 4)
    # number of lstm network pro tier
    n_rnn = 2
    # dimensionality of the lstms
    dim = 512

    # OPTIMIZATION

    # how many epochs should we train for
    max_epochs = 50
    # how many examples are used pro training steps
    batch_size = 16
    # the learning rate
    max_lr = 5e-4
    # betas control how fast the network changes its 'learning course'.
    # generally, betas should be close but smaller than 1. and be balanced with the batch_size :
    # the smaller the batch, the higher the betas 'could be'.
    betas = (0.9, 0.93)

    # MONITORING

    # how often should the network generate during training
    every_n_epochs = 4
    # how many examples from random prompts should be generated
    n_examples = 3
    # how many steps (1 step = 1 sample) should be generated
    n_steps = 15 * sr
    # the sampling temperature changes outputs a lot!
    # roughly : prefer values close to 1. & hot -> noisy ; cold -> silence
    temperature = torch.tensor([.9, .999, 1.25]).unsqueeze(1)

    assert temperature.size(0) == n_examples, "number of values in temperature must be equal to n_examples"
    print("arguments are ok!")

    """### create the data"""
    schema = mmk.SampleRNN.schema(sr, 0., q_levels)

    db_path = 'sample-rnn-demo.h5'
    print("collecting data...")
    db = mmk.Database.create(db_path, sources, schema)
    if not len(db.qx.files):
        raise ValueError("Empty db. No audio files were found...")
    print("successfully created the db.")

    """### create network and train"""
    net = mmk.SampleRNN(
        feature=schema['qx'],
        q_levels=q_levels,
        frame_sizes=frame_sizes,
        n_rnn=n_rnn,
        dim=dim,
        mlp_dim=dim,
        batch_size=batch_size,
        max_lr=max_lr,
        betas=betas,
        div_factor=5,
    )

    print(net.hparams)

    dm = mmk.DataModule(net, db,
                        splits=tuple(),
                        in_mem_data=True)

    cb = mmk.GenerateCallback(every_n_epochs, indices=[None] * n_examples,
                              n_steps=n_steps,
                              play_audios=True,
                              plot_audios=True,
                              temperature=temperature.to('cuda') if torch.cuda.is_available() else temperature)

    trainer = mmk.get_trainer(root_dir=None,
                              max_epochs=max_epochs,
                              callbacks=[cb],
                              checkpoint_callback=False)
    print("here we go!")
    trainer.fit(net, datamodule=dm)

    """----------------------------"""


if __name__ == '__main__':
    demo()



import dataclasses as dtc
import torch.nn as nn
from itertools import accumulate
import operator
import torch

from ..abstract.features import SegmentLabels, FilesLabels
from ..audios import MuLawSignal
from ..data import Feature, Input, AsSlice, Target
from . import model
from ..networks import WNNetwork
from .parts import IData, SuperAdam, SequenceModel

__all__ = [
    'WaveNetData',
    'WaveNet'
]


@dtc.dataclass
class WaveNetData(IData):
    feature: Feature = None
    batch_size: int = 32
    batch_seq_length: int = 64

    @classmethod
    def schema(cls, sr=22050, emphasis=0., q_levels=256,
               segment_labels=False, files_labels=False):

        schema = {"qx": MuLawSignal(sr=sr, emphasis=emphasis, q_levels=q_levels)}
        if segment_labels:
            # not yet supported...
            pass
            # schema.update({
            #     'loc': SegmentLabels(input_key='qx')
            # })
        if files_labels:
            schema.update({
                'glob': FilesLabels(input_key='qx')
            })
        return schema

    @classmethod
    def dependant_hp(cls, db):
        hp = {}
        if 'loc' in db.features:
            hp.update(dict(n_cin_classes=len(db.loc.regions)))
        if 'glob' in db.features:
            hp.update(dict(n_gin_classes=len(db.glob.files)))
        return dict(
            feature=db.schema['qx'], q_levels=db.schema['qx'].q_levels, **hp
        )

    def batch_signature(self, stage='fit'):
        inpt = [Input('qx', AsSlice(shift=0, length=self.batch_seq_length))]
        trgt = Target('qx', AsSlice(shift=self.shift,
                                    length=self.output_shape((-1, self.batch_seq_length, -1))[1]))
        # where are we conditioned?
        loc, glob = self.n_cin_classes is not None, self.n_gin_classes is not None
        if loc:
            pass
        if glob:
            pass
        if stage in ('full', 'fit', 'train', 'val'):
            return inpt, trgt
        # test, predict, generate...
        return inpt

    def loader_kwargs(self, stage, datamodule):
        return dict(
            batch_size=self.batch_size,
            drop_last=False,
            shuffle=True
        )


@model
class WaveNet(
    WaveNetData,
    SequenceModel,
    SuperAdam,
    WNNetwork,
):

    @staticmethod
    def loss_fn(output, target):
        criterion = nn.CrossEntropyLoss(reduction="mean")
        return {"loss": criterion(output.view(-1, output.size(-1)), target.view(-1))}

    def setup(self, stage: str):
        SequenceModel.setup(self, stage)
        SuperAdam.setup(self, stage)

    def encode_inputs(self, inputs: torch.Tensor):
        return self.feature.encode(inputs)

    def decode_outputs(self, outputs: torch.Tensor):
        return self.feature.decode(outputs)

    @staticmethod
    def rf(n_layers, kernel_size):
        if isinstance(kernel_size, tuple):
            assert sum(n_layers) == len(kernel_size), "total number of layers and of kernel sizes must match"
            k_iter = kernel_size
            dilations = list(accumulate([1, *kernel_size], operator.mul))
        else:
            # reverse_dilation_order leads to the connectivity of the FFTNet
            k_iter = [kernel_size] * sum(n_layers)
            dilations = [kernel_size ** (i)
                         for block in n_layers for i in range(block)]
        seq = list(dilations[i-1] * k_iter[i-1] for i in accumulate(n_layers))
        return sum(seq) - len(seq) + 1


def demo():
    """### import and arguments"""
    import mimikit as mmk
    import torch

    # DATA

    # list of files or directories to use as data ("./" is the cwd of the notebook)
    sources = ['./data']
    # audio sample rate
    sr = 16000
    # number of quantization levels (256 -> 8-bit)
    q_levels = 256

    # NETWORK

    # the number of layers determines 'how much past' is used to predict the next future step
    # here you can make blocks of layers by specifying a tuple of integers, e.g. (2, 3, 2)
    n_layers = (3,)
    # kernel_size is the size of the convolution. You can specify a single int for the whole
    # network or one size per layer
    kernel_size = (16, 8, 2)
    # how many parameters pro convolution layer
    gate_dim = 256
    # next arg can take 3 values : -1 -> input & output are summed at the end of the input,
    # 1 -> at the beginning, 0 -> they are not summed
    accum_outputs = 0
    # the next 2 args can take integers or None. Integers add skips and/or residuals layers of this size.
    # None adds no layers
    skip_dim = None
    residuals_dim = None

    # OPTIMIZATION

    # how many epochs should we train for
    max_epochs = 50
    # how many examples are used pro training steps
    batch_size = 16
    # the learning rate
    max_lr = 5e-4
    # betas control how fast the network changes its 'learning course'.
    # generally, betas should be close but smaller than 1. and be balanced with the batch_size :
    # the smaller the batch, the higher the betas 'could be'.
    betas = (0.9, 0.93)
    # one wavenet epoch can be very long, so as to monitor the net's progress,
    # we limit the number of batches pro epoch
    limit_train_batches = 1000

    # MONITORING

    # how often should the network generate during training
    every_n_epochs = 4
    # how many examples from random prompts should be generated
    n_examples = 3
    # how many steps (1 step = 1 sample) should be generated
    n_steps = 5 * sr
    # the sampling temperature changes outputs a lot!
    # roughly : prefer values close to 1. & hot -> noisy ; cold -> silence
    temperature = torch.tensor([.9, .999, 1.25]).unsqueeze(1)

    assert temperature.size(0) == n_examples, "number of values in temperature must be equal to n_examples"
    rf = mmk.WaveNet.rf(n_layers, kernel_size)
    print("arguments are ok! The network will have a receptive field of size :", rf, "samples")

    """### create the data"""
    schema = mmk.WaveNet.schema(sr, 0., q_levels)

    db_path = 'wavenet-demo.h5'
    print("collecting data...")
    db = mmk.Database.create(db_path, sources, schema)
    if not len(db.qx.files):
        raise ValueError("Empty db. No audio files were found...")
    print("successfully created the db.")

    """### create network and train"""
    net = mmk.WaveNet(
        **mmk.WaveNet.dependant_hp(db),
        kernel_size=kernel_size,
        gate_dim=gate_dim,
        accum_outputs=accum_outputs,
        residuals_dim=residuals_dim,
        skip_dim=skip_dim,
        n_layers=n_layers,
        batch_size=batch_size,
        batch_seq_length=rf * 2 if rf <= 128 else rf + rf // 4,
        max_lr=max_lr,
        betas=betas,
        div_factor=5,
    )
    print(net.hparams)

    dm = mmk.DataModule(net, db,
                        splits=tuple(),
                        in_mem_data=True)

    cb = mmk.GenerateCallback(every_n_epochs, indices=[None] * n_examples,
                              n_steps=n_steps,
                              play_audios=True,
                              plot_audios=True,
                              temperature=temperature.to('cuda') if torch.cuda.is_available() else temperature)

    trainer = mmk.get_trainer(root_dir=None,
                              max_epochs=max_epochs,
                              callbacks=[cb],
                              limit_train_batches=limit_train_batches,
                              checkpoint_callback=False)
    print("here we go!")
    trainer.fit(net, datamodule=dm)

    """----------------------------"""


if __name__ == '__main__':
    demo()



import numpy as np
import pytorch_lightning as pl
import torch
import torch.nn as nn
import dataclasses as dtc

from ..audios import Spectrogram
from ..data import Feature, AsSlice, Input, Target
from .parts import SuperAdam, SequenceModel, IData
from ..networks import Seq2SeqLSTM
from .parts.loss_functions import mean_L1_prop
from .model import model

__all__ = [
    'Seq2SeqLSTMModel'
]


@dtc.dataclass(init=True, repr=False, eq=False, frozen=False, unsafe_hash=True)
class Seq2SeqData(IData):
    feature: Feature = None
    batch_size: int = 16
    shift: int = 8

    @classmethod
    def schema(cls, sr=22050, emphasis=0., n_fft=2048, hop_length=512):
        schema = {"fft": Spectrogram(sr=sr, emphasis=emphasis,
                                     n_fft=n_fft, hop_length=hop_length,
                                     magspec=True)}
        return schema

    @classmethod
    def dependant_hp(cls, db):
        return dict(
            feature=Spectrogram(**db.fft.attrs), input_dim=db.fft.shape[-1]
        )

    def batch_signature(self, stage='fit'):
        inpt = Input('fft', AsSlice(shift=0, length=self.shift))
        trgt = Target('fft', AsSlice(shift=self.shift,
                                     length=self.shift))
        if stage in ('full', 'fit', 'train', 'val'):
            return inpt, trgt
        # test, predict, generate...
        return inpt

    def loader_kwargs(self, stage, datamodule):
        return dict(
            batch_size=self.batch_size,
            drop_last=False,
            shuffle=True
        )


@model
class Seq2SeqLSTMModel(
    Seq2SeqData,
    SuperAdam,
    SequenceModel,
    Seq2SeqLSTM,
):

    @staticmethod
    def loss_fn(output, target):
        return {"loss": mean_L1_prop(output, target)}

    def encode_inputs(self, inputs: torch.Tensor):
        return self.feature.encode(inputs)

    def decode_outputs(self, outputs: torch.Tensor):
        return self.feature.decode(outputs)


def demo():
    """### import and arguments"""
    import mimikit as mmk

    # DATA

    # list of files or directories to use as data ("./" is the cwd of the notebook)
    sources = ['./data']
    # audio sample rate
    sr = 22050
    # the size of the stft
    n_fft = 2048
    # hop_length of the
    hop_length = n_fft // 4

    # NETWORK

    # this network takes `shift` fft frames as input and outputs `shift` future frames
    shift = 8
    # the net contains at least 3 LSTM modules (1 in the Encoder, 2 in the Decoder)
    # you can add modules to the Encoder by increasing the next argument
    n_lstm = 1
    # all LSTM modules have internally the same number of layers :
    num_layers = 1
    # the dimensionality of the model
    model_dim = 1024

    # OPTIMIZATION

    # how many epochs should we train for
    max_epochs = 50
    # how many examples are used pro training steps
    batch_size = 16
    # the learning rate
    max_lr = 1e-3
    # betas control how fast the network changes its 'learning course'.
    # generally, betas should be close but smaller than 1. and be balanced with the batch_size :
    # the smaller the batch, the higher the betas 'could be'.
    betas = (0.9, 0.93)

    # MONITORING

    # how often should the network generate during training
    every_n_epochs = 2
    # how many examples from random prompts should be generated
    n_examples = 3
    # how many steps (1 step = `shift` fft frames!) should be generated
    n_steps = 1000 // shift

    print("arguments are ok!")

    """### create the data"""
    schema = mmk.Seq2SeqLSTMModel.schema(sr, n_fft=n_fft, hop_length=hop_length)

    db_path = 's2s-demo.h5'
    print("collecting data...")
    db = mmk.Database.create(db_path, sources, schema)
    if not len(db.fft.files):
        raise ValueError("Empty db. No audio files were found")
    print("successfully created the db.")

    """### create network and train"""
    net = mmk.Seq2SeqLSTMModel(
        **mmk.Seq2SeqLSTMModel.dependant_hp(db),
        shift=shift,
        n_lstm=n_lstm,
        num_layers=num_layers,
        model_dim=model_dim,
        batch_size=batch_size,
        max_lr=max_lr,
        div_factor=5,
        betas=betas,

    )
    print(net.hparams)

    dm = mmk.DataModule(net, db, splits=tuple())

    cb = mmk.GenerateCallback(every_n_epochs,
                              indices=[None] * n_examples,
                              n_steps=n_steps,
                              play_audios=True,
                              plot_audios=True)

    trainer = mmk.get_trainer(root_dir=None,
                              max_epochs=max_epochs,
                              callbacks=[cb],
                              checkpoint_callback=False)
    print("here we go!")
    trainer.fit(net, datamodule=dm)

    """----------------------------"""
